/******************************************************************************************
*	Chili DirectX Framework Version 16.07.20											  *
*	Graphics.h																			  *
*	Copyright 2016 PlanetChili <http://www.planetchili.net>								  *
*																						  *
*	This file is part of The Chili DirectX Framework.									  *
*																						  *
*	The Chili DirectX Framework is free software: you can redistribute it and/or modify	  *
*	it under the terms of the GNU General Public License as published by				  *
*	the Free Software Foundation, either version 3 of the License, or					  *
*	(at your option) any later version.													  *
*																						  *
*	The Chili DirectX Framework is distributed in the hope that it will be useful,		  *
*	but WITHOUT ANY WARRANTY; without even the implied warranty of						  *
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the						  *
*	GNU General Public License for more details.										  *
*																						  *
*	You should have received a copy of the GNU General Public License					  *
*	along with The Chili DirectX Framework.  If not, see <http://www.gnu.org/licenses/>.  *
******************************************************************************************/
#pragma once
#include "ChiliWin.h"
#include <d3d11.h>
#include <wrl.h>
#include "ChiliException.h"
#include "Colors.h"
#include "Surface.h"
#include "Rect.h"
#include <assert.h>

class Graphics
{
public:
	class Exception : public ChiliException
	{
	public:
		Exception( HRESULT hr,const std::wstring& note,const wchar_t* file,unsigned int line );
		std::wstring GetErrorName() const;
		std::wstring GetErrorDescription() const;
		virtual std::wstring GetFullMessage() const override;
		virtual std::wstring GetExceptionType() const override;
	private:
		HRESULT hr;
	};
private:
	// vertex format for the framebuffer fullscreen textured quad
	struct FSQVertex
	{
		float x,y,z;		// position
		float u,v;			// texcoords
	};
public:
	Graphics( class HWNDKey& key );
	Graphics( const Graphics& ) = delete;
	Graphics& operator=( const Graphics& ) = delete;
	void EndFrame();
	void BeginFrame();
	void PutPixel( int x,int y,int r,int g,int b )
	{
		PutPixel( x,y,{ unsigned char( r ),unsigned char( g ),unsigned char( b ) } );
	}
	void PutPixel( int x,int y,Color c );
	Color GetPixel(int x, int y) const;
	
	template<typename E>
	void DrawSprite(int x, int y, const Surface & surface, E effect, float scale = 1.0f)
	{
		DrawSprite(x, y, surface.GetRect(), surface, effect, scale);
	}

	template<typename E>
	void DrawSprite(int x, int y, const RectI & source, const Surface & surface, E effect, float scale = 1.0f)
	{
		DrawSprite(x, y, source, GetScreenRect(), surface, effect, scale);
	}

	template<typename E>
	void DrawSprite(int x, int y, RectI source, const RectI& clip, const Surface& surface, E effect, float scale = 1.0f)
	{
		assert(source.left >= 0);
		assert(source.right <= surface.GetWidth());
		assert(source.top >= 0);
		assert(source.bottom <= surface.GetHeight());
		assert(scale > 0.0f);

		const RectI scaleRect =
			RectI(
			x, 
			x + int(source.GetWidth() * scale),
			y,
			y + int(source.GetHeight() * scale)
			);

		for (int scaleY = scaleRect.top; scaleY < scaleRect.bottom; scaleY++)
		{
			for (int scaleX = scaleRect.left; scaleX < scaleRect.right; scaleX++)
			{
				const int sourceX = int((scaleX - scaleRect.left) / scale) + source.left;
				const int sourceY = int((scaleY - scaleRect.top) / scale) + source.top;

				if (clip.Contains({ scaleX, scaleY }))
				{
					effect(
						surface.GetPixel(sourceX, sourceY),
						scaleX,
						scaleY,
						*this
					);
				}
			}
		}

	}

	~Graphics();
private:
	Microsoft::WRL::ComPtr<IDXGISwapChain>				pSwapChain;
	Microsoft::WRL::ComPtr<ID3D11Device>				pDevice;
	Microsoft::WRL::ComPtr<ID3D11DeviceContext>			pImmediateContext;
	Microsoft::WRL::ComPtr<ID3D11RenderTargetView>		pRenderTargetView;
	Microsoft::WRL::ComPtr<ID3D11Texture2D>				pSysBufferTexture;
	Microsoft::WRL::ComPtr<ID3D11ShaderResourceView>	pSysBufferTextureView;
	Microsoft::WRL::ComPtr<ID3D11PixelShader>			pPixelShader;
	Microsoft::WRL::ComPtr<ID3D11VertexShader>			pVertexShader;
	Microsoft::WRL::ComPtr<ID3D11Buffer>				pVertexBuffer;
	Microsoft::WRL::ComPtr<ID3D11InputLayout>			pInputLayout;
	Microsoft::WRL::ComPtr<ID3D11SamplerState>			pSamplerState;
	D3D11_MAPPED_SUBRESOURCE							mappedSysBufferTexture;
	Color*                                              pSysBuffer = nullptr;
public:
	static constexpr int ScreenWidth = 800;
	static constexpr int ScreenHeight = 600;
	static RectI GetScreenRect();
};

#include "SpriteEffect.h"

#ifndef OPTIMIZE_SPRITE
extern template
void Graphics::DrawSprite<SpriteEffect::Copy>(int x, int y, RectI srcRect, 
	const RectI& clip, const Surface& s, SpriteEffect::Copy effect, float scale);

extern template
void Graphics::DrawSprite<SpriteEffect::Chroma>(int x, int y, RectI srcRect,
	const RectI& clip, const Surface& s, SpriteEffect::Chroma effect, float scale);

extern template
void Graphics::DrawSprite<SpriteEffect::Ghost>(int x, int y, RectI srcRect,
	const RectI& clip, const Surface& s, SpriteEffect::Ghost effect, float scale);

extern template
void Graphics::DrawSprite<SpriteEffect::Substitution>(int x, int y, RectI srcRect,
	const RectI& clip, const Surface& s, SpriteEffect::Substitution effect, float scale);


#endif // !OPTIMIZE_SPRITE


